/**
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */
'use strict';

var code = require('@lexical/code');
var list = require('@lexical/list');
var lexical = require('lexical');
var link = require('@lexical/link');
var richText = require('@lexical/rich-text');
var text = require('@lexical/text');

/**
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 */
/*
   How to add a new syntax to capture and transform.
   1. Create a new enumeration by adding to MarkdownFormatKind.
   2. Add a new criteria with a regEx pattern. See markdownStrikethrough as an example.
   3. Add your block criteria (e.g. '# ') to allMarkdownCriteria or 
      your text criteria (e.g. *MyItalic*) to allMarkdownCriteriaForTextNodes.
   4. Add your Lexical block specific transforming code here: transformTextNodeForText.   
      Add your Lexical text specific transforming code here: transformTextNodeForText.   
   */
// The trigger state helps to capture EditorState information
// from the prior and current EditorState.
// This is then used to determined if an auto format has been triggered.

// Eventually we need to support multiple trigger string's including newlines.
const SEPARATOR_BETWEEN_TEXT_AND_NON_TEXT_NODES = '\u0004'; // Select an unused unicode character to separate text and non-text nodes.

const SEPARATOR_LENGTH = SEPARATOR_BETWEEN_TEXT_AND_NON_TEXT_NODES.length;
const spaceTrigger = {
  triggerKind: 'space_trigger',
  triggerString: '\u0020'
}; // Future todo: add support for ``` + carriage return either inside or not inside code block. Should toggle between.
// const codeBlockTrigger : AutoFormatTrigger = {
//     triggerKind: 'codeBlock_trigger',
//     triggerString: '```', // + new paragraph element or new code block element.
// };

const triggers = [spaceTrigger
/*, codeBlockTrigger*/
]; // Future Todo: speed up performance by having non-capture group variations of the regex.

const autoFormatBase = {
  markdownFormatKind: null,
  regEx: /(?:)/,
  regExForAutoFormatting: /(?:)/,
  requiresParagraphStart: false
};
const paragraphStartBase = { ...autoFormatBase,
  requiresParagraphStart: true
};
const markdownHeader1 = { ...paragraphStartBase,
  markdownFormatKind: 'paragraphH1',
  regEx: /^(?:# )/,
  regExForAutoFormatting: /^(?:# )/
};
const markdownHeader2 = { ...paragraphStartBase,
  markdownFormatKind: 'paragraphH2',
  regEx: /^(?:## )/,
  regExForAutoFormatting: /^(?:## )/
};
const markdownHeader3 = { ...paragraphStartBase,
  markdownFormatKind: 'paragraphH3',
  regEx: /^(?:### )/,
  regExForAutoFormatting: /^(?:### )/
};
const markdownHeader4 = { ...paragraphStartBase,
  markdownFormatKind: 'paragraphH4',
  regEx: /^(?:#### )/,
  regExForAutoFormatting: /^(?:#### )/
};
const markdownHeader5 = { ...paragraphStartBase,
  markdownFormatKind: 'paragraphH5',
  regEx: /^(?:##### )/,
  regExForAutoFormatting: /^(?:##### )/
};
const markdownBlockQuote = { ...paragraphStartBase,
  markdownFormatKind: 'paragraphBlockQuote',
  regEx: /^(?:> )/,
  regExForAutoFormatting: /^(?:> )/
};
const markdownUnorderedListDash = { ...paragraphStartBase,
  markdownFormatKind: 'paragraphUnorderedList',
  regEx: /^(\s{0,10})(?:- )/,
  regExForAutoFormatting: /^(\s{0,10})(?:- )/
};
const markdownUnorderedListAsterisk = { ...paragraphStartBase,
  markdownFormatKind: 'paragraphUnorderedList',
  regEx: /^(\s{0,10})(?:\* )/,
  regExForAutoFormatting: /^(\s{0,10})(?:\* )/
};
const markdownCodeBlock = { ...paragraphStartBase,
  markdownFormatKind: 'paragraphCodeBlock',
  regEx: /^(```)$/,
  regExForAutoFormatting: /^(```)([a-z]*)( )/
};
const markdownOrderedList = { ...paragraphStartBase,
  markdownFormatKind: 'paragraphOrderedList',
  regEx: /^(\s{0,10})(\d+)\.\s/,
  regExForAutoFormatting: /^(\s{0,10})(\d+)\.\s/
};
const markdownHorizontalRule = { ...paragraphStartBase,
  markdownFormatKind: 'horizontalRule',
  regEx: /^(?:\*\*\*)$/,
  regExForAutoFormatting: /^(?:\*\*\* )/
};
const markdownHorizontalRuleUsingDashes = { ...paragraphStartBase,
  markdownFormatKind: 'horizontalRule',
  regEx: /^(?:---)$/,
  regExForAutoFormatting: /^(?:--- )/
};
const markdownInlineCode = { ...autoFormatBase,
  markdownFormatKind: 'code',
  regEx: /(`)([^`]*)(`)/,
  regExForAutoFormatting: /(`)(\s*\b)([^`]*)(\b\s*)(`)(\s)$/
};
const markdownBold = { ...autoFormatBase,
  markdownFormatKind: 'bold',
  regEx: /(\*\*)(\s*)([^\*\*]*)(\s*)(\*\*)()/,
  regExForAutoFormatting: /(\*\*)(\s*\b)([^\*\*]*)(\b\s*)(\*\*)(\s)$/
};
const markdownItalic = { ...autoFormatBase,
  markdownFormatKind: 'italic',
  regEx: /(\*)(\s*)([^\*]*)(\s*)(\*)()/,
  regExForAutoFormatting: /(\*)(\s*\b)([^\*]*)(\b\s*)(\*)(\s)$/
};
const markdownBold2 = { ...autoFormatBase,
  markdownFormatKind: 'bold',
  regEx: /(__)(\s*)([^__]*)(\s*)(__)()/,
  regExForAutoFormatting: /(__)(\s*)([^__]*)(\s*)(__)(\s)$/
};
const markdownItalic2 = { ...autoFormatBase,
  markdownFormatKind: 'italic',
  regEx: /(_)()([^_]*)()(_)()/,
  regExForAutoFormatting: /(_)()([^_]*)()(_)(\s)$/ // Maintain 7 groups.

}; // Markdown does not support underline, but we can allow folks to use
// the HTML tags for underline.

const fakeMarkdownUnderline = { ...autoFormatBase,
  markdownFormatKind: 'underline',
  regEx: /(\<u\>)(\s*)([^\<]*)(\s*)(\<\/u\>)()/,
  regExForAutoFormatting: /(\<u\>)(\s*\b)([^\<]*)(\b\s*)(\<\/u\>)(\s)$/
};
const markdownStrikethrough = { ...autoFormatBase,
  markdownFormatKind: 'strikethrough',
  regEx: /(~~)(\s*)([^~~]*)(\s*)(~~)()/,
  regExForAutoFormatting: /(~~)(\s*\b)([^~~]*)(\b\s*)(~~)(\s)$/
};
const markdownStrikethroughItalicBold = { ...autoFormatBase,
  markdownFormatKind: 'strikethrough_italic_bold',
  regEx: /(~~_\*\*)(\s*\b)([^~~_\*\*][^\*\*_~~]*)(\b\s*)(\*\*_~~)()/,
  regExForAutoFormatting: /(~~_\*\*)(\s*\b)([^~~_\*\*][^\*\*_~~]*)(\b\s*)(\*\*_~~)(\s)$/
};
const markdownItalicbold = { ...autoFormatBase,
  markdownFormatKind: 'italic_bold',
  regEx: /(_\*\*)(\s*\b)([^_\*\*][^\*\*_]*)(\b\s*)(\*\*_)/,
  regExForAutoFormatting: /(_\*\*)(\s*\b)([^_\*\*][^\*\*_]*)(\b\s*)(\*\*_)(\s)$/
};
const markdownStrikethroughItalic = { ...autoFormatBase,
  markdownFormatKind: 'strikethrough_italic',
  regEx: /(~~_)(\s*)([^~~_][^_~~]*)(\s*)(_~~)/,
  regExForAutoFormatting: /(~~_)(\s*)([^~~_][^_~~]*)(\s*)(_~~)(\s)$/
};
const markdownStrikethroughBold = { ...autoFormatBase,
  markdownFormatKind: 'strikethrough_bold',
  regEx: /(~~\*\*)(\s*\b)([^~~\*\*][^\*\*~~]*)(\b\s*)(\*\*~~)/,
  regExForAutoFormatting: /(~~\*\*)(\s*\b)([^~~\*\*][^\*\*~~]*)(\b\s*)(\*\*~~)(\s)$/
};
const markdownLink = { ...autoFormatBase,
  markdownFormatKind: 'link',
  regEx: /(\[)([^\]]*)(\]\()([^)]*)(\)*)()/,
  regExForAutoFormatting: /(\[)([^\]]*)(\]\()([^)]*)(\)*)(\s)$/
};
const allMarkdownCriteriaForTextNodes = [// Place the combination formats ahead of the individual formats.
// Combos
markdownStrikethroughItalicBold, markdownItalicbold, markdownStrikethroughItalic, markdownStrikethroughBold, // Individuals
markdownInlineCode, markdownBold, markdownItalic, // Must appear after markdownBold
markdownBold2, markdownItalic2, // Must appear after markdownBold2.
fakeMarkdownUnderline, markdownStrikethrough, markdownLink];
const allMarkdownCriteriaForParagraphs = [markdownHeader1, markdownHeader2, markdownHeader3, markdownHeader4, markdownHeader5, markdownBlockQuote, markdownUnorderedListDash, markdownUnorderedListAsterisk, markdownOrderedList, markdownCodeBlock, markdownHorizontalRule, markdownHorizontalRuleUsingDashes];
const allMarkdownCriteria = [...allMarkdownCriteriaForParagraphs, ...allMarkdownCriteriaForTextNodes];
function getAllTriggers() {
  return triggers;
}
function getAllMarkdownCriteriaForParagraphs() {
  return allMarkdownCriteriaForParagraphs;
}
function getAllMarkdownCriteriaForTextNodes() {
  return allMarkdownCriteriaForTextNodes;
}
function getAllMarkdownCriteria() {
  return allMarkdownCriteria;
}
function getInitialScanningContext(editor, isAutoFormatting, textNodeWithOffset, triggerState) {
  return {
    currentElementNode: null,
    editor,
    isAutoFormatting,
    isWithinCodeBlock: false,
    joinedText: null,
    markdownCriteria: {
      markdownFormatKind: 'noTransformation',
      regEx: /(?:)/,
      // Empty reg ex.
      regExForAutoFormatting: /(?:)/,
      // Empty reg ex.
      requiresParagraphStart: null
    },
    patternMatchResults: {
      regExCaptureGroups: []
    },
    textNodeWithOffset,
    triggerState
  };
}
function resetScanningContext(scanningContext) {
  scanningContext.joinedText = null;
  scanningContext.markdownCriteria = {
    markdownFormatKind: 'noTransformation',
    regEx: /(?:)/,
    // Empty reg ex.
    regExForAutoFormatting: /(?:)/,
    // Empty reg ex.
    requiresParagraphStart: null
  };
  scanningContext.patternMatchResults = {
    regExCaptureGroups: []
  };
  scanningContext.triggerState = null;
  scanningContext.textNodeWithOffset = null;
  return scanningContext;
}
function getCodeBlockCriteria() {
  return markdownCodeBlock;
}
function getPatternMatchResultsForCriteria(markdownCriteria, scanningContext, parentElementNode) {
  if (markdownCriteria.requiresParagraphStart === true) {
    return getPatternMatchResultsForParagraphs(markdownCriteria, scanningContext);
  }

  return getPatternMatchResultsForText(markdownCriteria, scanningContext, parentElementNode);
}
function getPatternMatchResultsForCodeBlock(scanningContext, text) {
  const markdownCriteria = getCodeBlockCriteria();
  return getPatternMatchResultsWithRegEx(text, true, false, scanningContext.isAutoFormatting ? markdownCriteria.regExForAutoFormatting : markdownCriteria.regEx);
}

function getPatternMatchResultsWithRegEx(textToSearch, matchMustAppearAtStartOfString, matchMustAppearAtEndOfString, regEx) {
  const patternMatchResults = {
    regExCaptureGroups: []
  };
  const regExMatches = textToSearch.match(regEx);

  if (regExMatches !== null && regExMatches.length > 0 && (matchMustAppearAtStartOfString === false || regExMatches.index === 0) && (matchMustAppearAtEndOfString === false || regExMatches.index + regExMatches[0].length === textToSearch.length)) {
    const captureGroupsCount = regExMatches.length;
    let runningLength = regExMatches.index;

    for (let captureGroupIndex = 0; captureGroupIndex < captureGroupsCount; captureGroupIndex++) {
      const textContent = regExMatches[captureGroupIndex];
      patternMatchResults.regExCaptureGroups.push({
        offsetInParent: runningLength,
        text: textContent
      }); // The 0th capture group is special in that it's text contents is
      // a join of all subsequent capture groups. So, skip this group
      // when calculating the runningLength.

      if (captureGroupIndex > 0) {
        runningLength += textContent.length;
      }
    }

    return patternMatchResults;
  }

  return null;
}

function hasPatternMatchResults(scanningContext) {
  return scanningContext.patternMatchResults.regExCaptureGroups.length > 0;
}
function getTextNodeWithOffsetOrThrow(scanningContext) {
  const textNodeWithOffset = scanningContext.textNodeWithOffset;

  if (textNodeWithOffset == null) {
    {
      throw Error(`Expect to have a text node with offset.`);
    }
  }

  return textNodeWithOffset;
}

function getPatternMatchResultsForParagraphs(markdownCriteria, scanningContext) {
  const textNodeWithOffset = getTextNodeWithOffsetOrThrow(scanningContext); // At start of paragraph.

  if (textNodeWithOffset.node.getPreviousSibling() === null) {
    const textToSearch = textNodeWithOffset.node.getTextContent();
    return getPatternMatchResultsWithRegEx(textToSearch, true, false, scanningContext.isAutoFormatting ? markdownCriteria.regExForAutoFormatting : markdownCriteria.regEx);
  }

  return null;
}

function getPatternMatchResultsForText(markdownCriteria, scanningContext, parentElementNode) {
  if (scanningContext.joinedText == null) {
    if (lexical.$isElementNode(parentElementNode)) {
      if (scanningContext.joinedText == null) {
        // Lazy calculate the text to search.
        scanningContext.joinedText = text.$joinTextNodesInElementNode(parentElementNode, SEPARATOR_BETWEEN_TEXT_AND_NON_TEXT_NODES, getTextNodeWithOffsetOrThrow(scanningContext));
      }
    } else {
      {
        throw Error(`Expected node ${parentElementNode.__key} to to be a ElementNode.`);
      }
    }
  }

  const matchMustAppearAtEndOfString = markdownCriteria.regExForAutoFormatting === true;
  return getPatternMatchResultsWithRegEx(scanningContext.joinedText, false, matchMustAppearAtEndOfString, scanningContext.isAutoFormatting ? markdownCriteria.regExForAutoFormatting : markdownCriteria.regEx);
}

function getNewNodeForCriteria(scanningContext, element, createHorizontalRuleNode) {
  let newNode = null;
  const shouldDelete = false;
  const children = element.getChildren();
  const markdownCriteria = scanningContext.markdownCriteria;
  const patternMatchResults = scanningContext.patternMatchResults;

  if (markdownCriteria.markdownFormatKind != null) {
    switch (markdownCriteria.markdownFormatKind) {
      case 'paragraphH1':
        {
          newNode = richText.$createHeadingNode('h1');
          newNode.append(...children);
          return {
            newNode,
            shouldDelete
          };
        }

      case 'paragraphH2':
        {
          newNode = richText.$createHeadingNode('h2');
          newNode.append(...children);
          return {
            newNode,
            shouldDelete
          };
        }

      case 'paragraphH3':
        {
          newNode = richText.$createHeadingNode('h3');
          newNode.append(...children);
          return {
            newNode,
            shouldDelete
          };
        }

      case 'paragraphH4':
        {
          newNode = richText.$createHeadingNode('h4');
          newNode.append(...children);
          return {
            newNode,
            shouldDelete
          };
        }

      case 'paragraphH5':
        {
          newNode = richText.$createHeadingNode('h5');
          newNode.append(...children);
          return {
            newNode,
            shouldDelete
          };
        }

      case 'paragraphBlockQuote':
        {
          newNode = richText.$createQuoteNode();
          newNode.append(...children);
          return {
            newNode,
            shouldDelete
          };
        }

      case 'paragraphUnorderedList':
        {
          createListOrMergeWithPrevious(element, children, patternMatchResults, 'ul');
          return {
            newNode: null,
            shouldDelete: false
          };
        }

      case 'paragraphOrderedList':
        {
          const startAsString = patternMatchResults.regExCaptureGroups.length > 1 ? patternMatchResults.regExCaptureGroups[patternMatchResults.regExCaptureGroups.length - 1].text : '1'; // For conversion, don't use start number.
          // For short-cuts aka autoFormatting, use start number.
          // Later, this should be surface dependent and externalized.

          const start = scanningContext.isAutoFormatting ? parseInt(startAsString, 10) : undefined;
          createListOrMergeWithPrevious(element, children, patternMatchResults, 'ol', start);
          return {
            newNode: null,
            shouldDelete: false
          };
        }

      case 'paragraphCodeBlock':
        {
          // Toggle code and paragraph nodes.
          if (scanningContext.isAutoFormatting === false) {
            const shouldToggle = hasPatternMatchResults(scanningContext);

            if (shouldToggle) {
              scanningContext.isWithinCodeBlock = scanningContext.isWithinCodeBlock !== true; // When toggling, always clear the code block element node.

              scanningContext.currentElementNode = null;
              return {
                newNode: null,
                shouldDelete: true
              };
            }

            if (scanningContext.isWithinCodeBlock) {
              // Create the code block and return it to the caller.
              if (scanningContext.currentElementNode == null) {
                const newCodeBlockNode = code.$createCodeNode();
                newCodeBlockNode.append(...children);
                scanningContext.currentElementNode = newCodeBlockNode;
                return {
                  newNode: newCodeBlockNode,
                  shouldDelete: false
                };
              } // Build up the code block with a line break and the children.


              if (scanningContext.currentElementNode != null) {
                const codeBlockNode = scanningContext.currentElementNode;
                const lineBreakNode = lexical.$createLineBreakNode();
                codeBlockNode.append(lineBreakNode);

                if (children.length) {
                  codeBlockNode.append(lineBreakNode);
                }

                codeBlockNode.append(...children);
              }
            }

            return {
              newNode: null,
              shouldDelete: true
            };
          }

          if (scanningContext.triggerState != null && scanningContext.triggerState.isCodeBlock) {
            newNode = lexical.$createParagraphNode();
          } else {
            newNode = code.$createCodeNode();
            const codingLanguage = patternMatchResults.regExCaptureGroups.length >= 3 ? patternMatchResults.regExCaptureGroups[2].text : null;

            if (codingLanguage != null && codingLanguage.length > 0) {
              newNode.setLanguage(codingLanguage);
            }
          }

          newNode.append(...children);
          return {
            newNode,
            shouldDelete
          };
        }

      case 'horizontalRule':
        {
          if (createHorizontalRuleNode != null) {
            // return null for newNode. Insert the HR here.
            const horizontalRuleNode = createHorizontalRuleNode();
            element.insertBefore(horizontalRuleNode);
          }

          break;
        }
    }
  }

  return {
    newNode,
    shouldDelete
  };
}

function createListOrMergeWithPrevious(element, children, patternMatchResults, tag, start) {
  const listItem = list.$createListItemNode();
  const indentMatch = patternMatchResults.regExCaptureGroups[0].text.match(/^\s*/);
  const indent = indentMatch ? Math.floor(indentMatch[0].length / 4) : 0;
  listItem.append(...children); // Checking if previous element is a list, and if so append
  // new list item inside instead of creating new list

  const prevElement = element.getPreviousSibling();

  if (list.$isListNode(prevElement) && prevElement.getTag() === tag) {
    prevElement.append(listItem);
    element.remove();
  } else {
    const list$1 = list.$createListNode(tag, start);
    list$1.append(listItem);
    element.replace(list$1);
  }

  if (indent) {
    listItem.setIndent(indent);
  }
}

function transformTextNodeForMarkdownCriteria(scanningContext, elementNode, createHorizontalRuleNode) {
  if (scanningContext.markdownCriteria.requiresParagraphStart === true) {
    transformTextNodeForElementNode(elementNode, scanningContext, createHorizontalRuleNode);
  } else {
    transformTextNodeForText(scanningContext, elementNode);
  }
}

function transformTextNodeForElementNode(elementNode, scanningContext, createHorizontalRuleNode) {
  if (scanningContext.textNodeWithOffset != null) {
    const textNodeWithOffset = getTextNodeWithOffsetOrThrow(scanningContext);

    if (hasPatternMatchResults(scanningContext)) {
      const text = scanningContext.patternMatchResults.regExCaptureGroups[0].text; // Remove the text which we matched.

      const textNode = textNodeWithOffset.node.spliceText(0, text.length, '', true);

      if (textNode.getTextContent() === '') {
        textNode.selectPrevious();
        textNode.remove();
      }
    }
  } // Transform the current element kind to the new element kind.


  const {
    newNode,
    shouldDelete
  } = getNewNodeForCriteria(scanningContext, elementNode, createHorizontalRuleNode);

  if (shouldDelete) {
    elementNode.remove();
  } else if (newNode !== null) {
    elementNode.replace(newNode);
  }
}

function transformTextNodeForText(scanningContext, parentElementNode) {
  const markdownCriteria = scanningContext.markdownCriteria;

  if (markdownCriteria.markdownFormatKind != null) {
    const formatting = getTextFormatType(markdownCriteria.markdownFormatKind);

    if (formatting != null) {
      transformTextNodeWithFormatting(formatting, scanningContext, parentElementNode);
      return;
    }

    if (markdownCriteria.markdownFormatKind === 'link') {
      transformTextNodeWithLink(scanningContext, parentElementNode);
    }
  }
}

function transformTextNodeWithFormatting(formatting, scanningContext, parentElementNode) {
  const patternMatchResults = scanningContext.patternMatchResults;
  const groupCount = patternMatchResults.regExCaptureGroups.length;

  if (groupCount !== 7) {
    // For BIUS and similar formats which have a pattern + text + pattern:
    // given '*italic* ' below are the capture groups by index:
    // 0. '*italic* '
    // 1. '*'
    // 2. whitespace  // typically this is "".
    // 3. 'italic'
    // 4. whitespace  // typicallly this is "".
    // 5. '*'
    // 6. ' '
    return;
  } // Remove unwanted text in reg ex pattern.
  // Remove group 5.


  removeTextByCaptureGroups(5, 5, scanningContext, parentElementNode); // Remove group 1.

  removeTextByCaptureGroups(1, 1, scanningContext, parentElementNode); // Apply the formatting.

  formatTextInCaptureGroupIndex(formatting, 3, scanningContext, parentElementNode); // Place caret at end of final capture group.

  selectAfterFinalCaptureGroup(scanningContext, parentElementNode);
}

function transformTextNodeWithLink(scanningContext, parentElementNode) {
  const patternMatchResults = scanningContext.patternMatchResults;
  const regExCaptureGroups = patternMatchResults.regExCaptureGroups;
  const groupCount = regExCaptureGroups.length;

  if (groupCount !== 7) {
    // For links and similar formats which have: pattern + text + pattern + pattern2 text2 + pattern2:
    // Given '[title](url) ', below are the capture groups by index:
    // 0. '[title](url) '
    // 1. '['
    // 2. 'title'
    // 3. ']('
    // 4. 'url'
    // 5. ')'
    // 6. ' '
    return;
  }

  const title = regExCaptureGroups[2].text;
  const url = regExCaptureGroups[4].text;

  if (title.length === 0 || url.length === 0) {
    return;
  } // Remove the initial pattern through to the final pattern.


  removeTextByCaptureGroups(1, 5, scanningContext, parentElementNode);
  insertTextPriorToCaptureGroup(1, // Insert at the beginning of the meaningful capture groups, namely index 1. Index 0 refers to the whole matched string.
  title, scanningContext, parentElementNode);
  const newSelectionForLink = createSelectionWithCaptureGroups(1, 1, false, true, scanningContext, parentElementNode);

  if (newSelectionForLink == null) {
    return;
  }

  lexical.$setSelection(newSelectionForLink);
  scanningContext.editor.dispatchCommand(link.TOGGLE_LINK_COMMAND, url); // Place caret at end of final capture group.

  selectAfterFinalCaptureGroup(scanningContext, parentElementNode);
} // Below are lower level helper functions.


function getParentElementNodeOrThrow(scanningContext) {
  return getTextNodeWithOffsetOrThrow(scanningContext).node.getParentOrThrow();
}

function getJoinedTextLength(patternMatchResults) {
  const groupCount = patternMatchResults.regExCaptureGroups.length;

  if (groupCount < 2) {
    // Ignore capture group 0, as regEx defaults the 0th one to the entire matched string.
    return 0;
  }

  const lastGroupIndex = groupCount - 1;
  return patternMatchResults.regExCaptureGroups[lastGroupIndex].offsetInParent + patternMatchResults.regExCaptureGroups[lastGroupIndex].text.length;
}

function getTextFormatType(markdownFormatKind) {
  switch (markdownFormatKind) {
    case 'italic':
    case 'bold':
    case 'underline':
    case 'strikethrough':
    case 'code':
      return [markdownFormatKind];

    case 'strikethrough_italic_bold':
      {
        return ['strikethrough', 'italic', 'bold'];
      }

    case 'italic_bold':
      {
        return ['italic', 'bold'];
      }

    case 'strikethrough_italic':
      {
        return ['strikethrough', 'italic'];
      }

    case 'strikethrough_bold':
      {
        return ['strikethrough', 'bold'];
      }
  }

  return null;
}

function createSelectionWithCaptureGroups(anchorCaptureGroupIndex, focusCaptureGroupIndex, startAtEndOfAnchor, finishAtEndOfFocus, scanningContext, parentElementNode) {
  const patternMatchResults = scanningContext.patternMatchResults;
  const regExCaptureGroups = patternMatchResults.regExCaptureGroups;
  const regExCaptureGroupsCount = regExCaptureGroups.length;

  if (anchorCaptureGroupIndex >= regExCaptureGroupsCount || focusCaptureGroupIndex >= regExCaptureGroupsCount) {
    return null;
  }

  const joinedTextLength = getJoinedTextLength(patternMatchResults);
  const anchorCaptureGroupDetail = regExCaptureGroups[anchorCaptureGroupIndex];
  const focusCaptureGroupDetail = regExCaptureGroups[focusCaptureGroupIndex];
  const anchorLocation = startAtEndOfAnchor ? anchorCaptureGroupDetail.offsetInParent + anchorCaptureGroupDetail.text.length : anchorCaptureGroupDetail.offsetInParent;
  const focusLocation = finishAtEndOfFocus ? focusCaptureGroupDetail.offsetInParent + focusCaptureGroupDetail.text.length : focusCaptureGroupDetail.offsetInParent;
  const anchorTextNodeWithOffset = text.$findNodeWithOffsetFromJoinedText(anchorLocation, joinedTextLength, SEPARATOR_LENGTH, parentElementNode);
  const focusTextNodeWithOffset = text.$findNodeWithOffsetFromJoinedText(focusLocation, joinedTextLength, SEPARATOR_LENGTH, parentElementNode);

  if (anchorTextNodeWithOffset == null && focusTextNodeWithOffset == null && parentElementNode.getChildren().length === 0) {
    const emptyElementSelection = lexical.$createRangeSelection();
    emptyElementSelection.anchor.set(parentElementNode.getKey(), 0, 'element');
    emptyElementSelection.focus.set(parentElementNode.getKey(), 0, 'element');
    return emptyElementSelection;
  }

  if (anchorTextNodeWithOffset == null || focusTextNodeWithOffset == null) {
    return null;
  }

  const selection = lexical.$createRangeSelection();
  selection.anchor.set(anchorTextNodeWithOffset.node.getKey(), anchorTextNodeWithOffset.offset, 'text');
  selection.focus.set(focusTextNodeWithOffset.node.getKey(), focusTextNodeWithOffset.offset, 'text');
  return selection;
}

function removeTextByCaptureGroups(anchorCaptureGroupIndex, focusCaptureGroupIndex, scanningContext, parentElementNode) {
  const patternMatchResults = scanningContext.patternMatchResults;
  const regExCaptureGroups = patternMatchResults.regExCaptureGroups;
  const newSelection = createSelectionWithCaptureGroups(anchorCaptureGroupIndex, focusCaptureGroupIndex, false, true, scanningContext, parentElementNode);

  if (newSelection != null) {
    lexical.$setSelection(newSelection);
    const currentSelection = lexical.$getSelection();

    if (currentSelection != null && lexical.$isRangeSelection(currentSelection) && currentSelection.isCollapsed() === false) {
      currentSelection.removeText(); // Shift all group offsets and clear out group text.

      let runningLength = 0;
      const groupCount = regExCaptureGroups.length;

      for (let i = anchorCaptureGroupIndex; i < groupCount; i++) {
        const captureGroupDetail = regExCaptureGroups[i];

        if (i > anchorCaptureGroupIndex) {
          captureGroupDetail.offsetInParent -= runningLength;
        }

        if (i <= focusCaptureGroupIndex) {
          runningLength += captureGroupDetail.text.length;
          captureGroupDetail.text = '';
        }
      }
    }
  }
}

function insertTextPriorToCaptureGroup(captureGroupIndex, text, scanningContext, parentElementNode) {
  const patternMatchResults = scanningContext.patternMatchResults;
  const regExCaptureGroups = patternMatchResults.regExCaptureGroups;
  const regExCaptureGroupsCount = regExCaptureGroups.length;

  if (captureGroupIndex >= regExCaptureGroupsCount) {
    return;
  }

  const captureGroupDetail = regExCaptureGroups[captureGroupIndex];
  const newCaptureGroupDetail = {
    offsetInParent: captureGroupDetail.offsetInParent,
    text
  };
  const newSelection = createSelectionWithCaptureGroups(captureGroupIndex, captureGroupIndex, false, false, scanningContext, parentElementNode);

  if (newSelection != null) {
    lexical.$setSelection(newSelection);
    const currentSelection = lexical.$getSelection();

    if (currentSelection != null && lexical.$isRangeSelection(currentSelection) && currentSelection.isCollapsed()) {
      currentSelection.insertText(newCaptureGroupDetail.text); // Update the capture groups.

      regExCaptureGroups.splice(captureGroupIndex, 0, newCaptureGroupDetail);
      const textLength = newCaptureGroupDetail.text.length;
      const newGroupCount = regExCaptureGroups.length;

      for (let i = captureGroupIndex + 1; i < newGroupCount; i++) {
        const currentCaptureGroupDetail = regExCaptureGroups[i];
        currentCaptureGroupDetail.offsetInParent += textLength;
      }
    }
  }
}

function formatTextInCaptureGroupIndex(formatTypes, captureGroupIndex, scanningContext, parentElementNode) {
  const patternMatchResults = scanningContext.patternMatchResults;
  const regExCaptureGroups = patternMatchResults.regExCaptureGroups;
  const regExCaptureGroupsCount = regExCaptureGroups.length;

  if (!(captureGroupIndex < regExCaptureGroupsCount)) {
    throw Error(`The capture group count in the RegEx does match the actual capture group count.`);
  }

  const captureGroupDetail = regExCaptureGroups[captureGroupIndex];

  if (captureGroupDetail.text.length === 0) {
    return;
  }

  const newSelection = createSelectionWithCaptureGroups(captureGroupIndex, captureGroupIndex, false, true, scanningContext, parentElementNode);

  if (newSelection != null) {
    lexical.$setSelection(newSelection);
    const currentSelection = lexical.$getSelection();

    if (lexical.$isRangeSelection(currentSelection)) {
      for (let i = 0; i < formatTypes.length; i++) {
        currentSelection.formatText(formatTypes[i]);
      }
    }
  }
} // Place caret at end of final capture group.


function selectAfterFinalCaptureGroup(scanningContext, parentElementNode) {
  const patternMatchResults = scanningContext.patternMatchResults;
  const groupCount = patternMatchResults.regExCaptureGroups.length;

  if (groupCount < 2) {
    // Ignore capture group 0, as regEx defaults the 0th one to the entire matched string.
    return;
  }

  const lastGroupIndex = groupCount - 1;
  const newSelection = createSelectionWithCaptureGroups(lastGroupIndex, lastGroupIndex, true, true, scanningContext, parentElementNode);

  if (newSelection != null) {
    lexical.$setSelection(newSelection);
  }
}

/**
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 */

function getTextNodeForAutoFormatting(selection) {
  if (!lexical.$isRangeSelection(selection)) {
    return null;
  }

  const node = selection.anchor.getNode();

  if (!lexical.$isTextNode(node)) {
    return null;
  }

  return {
    node,
    offset: selection.anchor.offset
  };
}

function updateAutoFormatting(editor, scanningContext, createHorizontalRuleNode) {
  editor.update(() => {
    const elementNode = getTextNodeWithOffsetOrThrow(scanningContext).node.getParentOrThrow();
    transformTextNodeForMarkdownCriteria(scanningContext, elementNode, createHorizontalRuleNode);
  }, {
    tag: 'history-push'
  });
}

function getCriteriaWithPatternMatchResults(markdownCriteriaArray, scanningContext) {
  const currentTriggerState = scanningContext.triggerState;
  const count = markdownCriteriaArray.length;

  for (let i = 0; i < count; i++) {
    const markdownCriteria = markdownCriteriaArray[i]; // Skip code block nodes, unless the autoFormatKind calls for toggling the code block.

    if (currentTriggerState != null && currentTriggerState.isCodeBlock === false || markdownCriteria.markdownFormatKind === 'paragraphCodeBlock') {
      const patternMatchResults = getPatternMatchResultsForCriteria(markdownCriteria, scanningContext, getParentElementNodeOrThrow(scanningContext));

      if (patternMatchResults != null) {
        return {
          markdownCriteria,
          patternMatchResults
        };
      }
    }
  }

  return {
    markdownCriteria: null,
    patternMatchResults: null
  };
}

function findScanningContextWithValidMatch(editor, currentTriggerState) {
  let scanningContext = null;
  editor.getEditorState().read(() => {
    const textNodeWithOffset = getTextNodeForAutoFormatting(lexical.$getSelection());

    if (textNodeWithOffset === null) {
      return;
    } // Please see the declaration of ScanningContext for a detailed explanation.


    const initialScanningContext = getInitialScanningContext(editor, true, textNodeWithOffset, currentTriggerState);
    const criteriaWithPatternMatchResults = getCriteriaWithPatternMatchResults( // Do not apply paragraph node changes like blockQuote or H1 to listNodes. Also, do not attempt to transform a list into a list using * or -.
    currentTriggerState.isParentAListItemNode === false ? getAllMarkdownCriteria() : getAllMarkdownCriteriaForTextNodes(), initialScanningContext);

    if (criteriaWithPatternMatchResults.markdownCriteria === null || criteriaWithPatternMatchResults.patternMatchResults === null) {
      return;
    }

    scanningContext = initialScanningContext; // Lazy fill-in the particular format criteria and any matching result information.

    scanningContext.markdownCriteria = criteriaWithPatternMatchResults.markdownCriteria;
    scanningContext.patternMatchResults = criteriaWithPatternMatchResults.patternMatchResults;
  });
  return scanningContext;
}

function getTriggerState(editorState) {
  let criteria = null;
  editorState.read(() => {
    const selection = lexical.$getSelection();

    if (!lexical.$isRangeSelection(selection) || !selection.isCollapsed()) {
      return;
    }

    const node = selection.anchor.getNode();
    const parentNode = node.getParent();
    const isParentAListItemNode = list.$isListItemNode(parentNode);
    const hasParentNode = parentNode !== null;
    criteria = {
      anchorOffset: selection.anchor.offset,
      hasParentNode,
      isCodeBlock: code.$isCodeNode(node),
      isParentAListItemNode,
      isSelectionCollapsed: true,
      isSimpleText: lexical.$isTextNode(node) && node.isSimpleText(),
      nodeKey: node.getKey(),
      textContent: node.getTextContent()
    };
  });
  return criteria;
}
function findScanningContext(editor, currentTriggerState, priorTriggerState) {
  if (currentTriggerState == null || priorTriggerState == null) {
    return null;
  }

  const triggerArray = getAllTriggers();
  const triggerCount = triggers.length;

  for (let ti = 0; ti < triggerCount; ti++) {
    const triggerString = triggerArray[ti].triggerString; // The below checks needs to execute relativey quickly, so perform the light-weight ones first.
    // The substr check is a quick way to avoid autoformat parsing in that it looks for the autoformat
    // trigger which is the trigger string (" ").

    const triggerStringLength = triggerString.length;
    const currentTextContentLength = currentTriggerState.textContent.length;
    const triggerOffset = currentTriggerState.anchorOffset - triggerStringLength; // Todo: these checks help w/ performance, yet we can do more.
    // We might consider looking for ** + space or __ + space and so on to boost performance
    // even further. Make sure the patter is driven from the trigger state type.

    if ((currentTriggerState.hasParentNode === true && currentTriggerState.isSimpleText && currentTriggerState.isSelectionCollapsed && currentTriggerState.anchorOffset !== priorTriggerState.anchorOffset && triggerOffset >= 0 && triggerOffset + triggerStringLength <= currentTextContentLength && currentTriggerState.textContent.substr(triggerOffset, triggerStringLength) === triggerString && // Some code differentiation needed if trigger kind is not a simple space character.
    currentTriggerState.textContent !== priorTriggerState.textContent) === false) {
      return null;
    }
  }

  return findScanningContextWithValidMatch(editor, currentTriggerState);
}

/**
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 */
function convertStringToLexical(text, editor) {
  if (!text.length) {
    return null;
  }

  const nodes = [];
  const splitLines = text.split('\n');
  const splitLinesCount = splitLines.length;

  for (let i = 0; i < splitLinesCount; i++) {
    if (splitLines[i].length > 0) {
      nodes.push(lexical.$createParagraphNode().append(lexical.$createTextNode(splitLines[i])));
    } else {
      nodes.push(lexical.$createParagraphNode());
    }
  }

  if (nodes.length) {
    const root = lexical.$getRoot();
    root.clear();
    root.append(...nodes);
    return root;
  }

  return null;
}
function convertMarkdownForElementNodes(editor, createHorizontalRuleNode) {
  // Please see the declaration of ScanningContext for a detailed explanation.
  const scanningContext = getInitialScanningContext(editor, false, null, null);
  const root = lexical.$getRoot();
  let done = false;
  let startIndex = 0; // Handle the paragraph level markdown.

  while (!done) {
    done = true;
    const elementNodes = root.getChildren();
    const countOfElementNodes = elementNodes.length;

    for (let i = startIndex; i < countOfElementNodes; i++) {
      const elementNode = elementNodes[i];

      if (lexical.$isElementNode(elementNode)) {
        convertParagraphLevelMarkdown(scanningContext, elementNode, createHorizontalRuleNode);
      } // Reset the scanning information that relates to the particular element node.


      resetScanningContext(scanningContext);

      if (root.getChildren().length !== countOfElementNodes) {
        // The conversion added or removed an from root's children.
        startIndex = i;
        done = false;
        break;
      }
    }
  } // while


  done = false;
  startIndex = 0; // Handle the text level markdown.

  while (!done) {
    done = true;
    const elementNodes = root.getChildren();
    const countOfElementNodes = elementNodes.length;

    for (let i = startIndex; i < countOfElementNodes; i++) {
      const elementNode = elementNodes[i];

      if (lexical.$isElementNode(elementNode)) {
        convertTextLevelMarkdown(scanningContext, elementNode, createHorizontalRuleNode);
      } // Reset the scanning information that relates to the particular element node.


      resetScanningContext(scanningContext);
    }
  } // while

}

function convertParagraphLevelMarkdown(scanningContext, elementNode, createHorizontalRuleNode) {
  const textContent = elementNode.getTextContent(); // Handle paragraph nodes below.

  if (lexical.$isParagraphNode(elementNode)) {
    const paragraphNode = elementNode;
    const firstChild = paragraphNode.getFirstChild();
    const firstChildIsTextNode = lexical.$isTextNode(firstChild); // Handle conversion to code block.

    if (scanningContext.isWithinCodeBlock === true) {
      if (firstChild != null && firstChildIsTextNode) {
        // Test if we encounter ending code block.
        scanningContext.textNodeWithOffset = {
          node: firstChild,
          offset: 0
        };
        const patternMatchResults = getPatternMatchResultsForCodeBlock(scanningContext, textContent);

        if (patternMatchResults != null) {
          // Toggle transform to or from code block.
          scanningContext.patternMatchResults = patternMatchResults;
        }
      }

      scanningContext.markdownCriteria = getCodeBlockCriteria(); // Perform text transformation here.

      transformTextNodeForMarkdownCriteria(scanningContext, elementNode, createHorizontalRuleNode);
      return;
    }

    if (elementNode.getChildren().length) {
      const allCriteria = getAllMarkdownCriteriaForParagraphs();
      const count = allCriteria.length;
      scanningContext.joinedText = paragraphNode.getTextContent();

      if (!(firstChild != null && firstChildIsTextNode)) {
        throw Error(`Expect paragraph containing only text nodes.`);
      }

      scanningContext.textNodeWithOffset = {
        node: firstChild,
        offset: 0
      };

      for (let i = 0; i < count; i++) {
        const criteria = allCriteria[i];

        if (criteria.requiresParagraphStart === false) {
          return;
        }

        const patternMatchResults = getPatternMatchResultsForCriteria(criteria, scanningContext, getParentElementNodeOrThrow(scanningContext));

        if (patternMatchResults != null) {
          scanningContext.markdownCriteria = criteria;
          scanningContext.patternMatchResults = patternMatchResults; // Perform text transformation here.

          transformTextNodeForMarkdownCriteria(scanningContext, elementNode, createHorizontalRuleNode);
          return;
        }
      }
    }
  }
}

function convertTextLevelMarkdown(scanningContext, elementNode, createHorizontalRuleNode) {
  const firstChild = elementNode.getFirstChild();

  if (lexical.$isTextNode(firstChild)) {
    // This function will convert all text nodes within the elementNode.
    convertMarkdownForTextCriteria(scanningContext, elementNode, createHorizontalRuleNode);
    return;
  } // Handle the case where the elementNode has child elementNodes like lists.
  // Since we started at a text import, we don't need to worry about anything but textNodes.


  const children = elementNode.getChildren();
  const countOfChildren = children.length;

  for (let i = 0; i < countOfChildren; i++) {
    const node = children[i];

    if (lexical.$isElementNode(node)) {
      // Recurse down until we find a text node.
      convertTextLevelMarkdown(scanningContext, node, createHorizontalRuleNode);
    }
  }
}

function convertMarkdownForTextCriteria(scanningContext, elementNode, createHorizontalRuleNode) {
  resetScanningContext(scanningContext); // Cycle through all the criteria and convert all text patterns in the parent element.

  const allCriteria = getAllMarkdownCriteriaForTextNodes();
  const count = allCriteria.length;
  let textContent = elementNode.getTextContent();
  let done = textContent.length === 0;
  let startIndex = 0;

  while (!done) {
    done = true;

    for (let i = startIndex; i < count; i++) {
      const criteria = allCriteria[i];

      if (scanningContext.textNodeWithOffset == null) {
        // Need to search through the very last text node in the element.
        const lastTextNode = getLastTextNodeInElementNode(elementNode);

        if (lastTextNode == null) {
          // If we have no more text nodes, then there's nothing to search and transform.
          return;
        }

        scanningContext.textNodeWithOffset = {
          node: lastTextNode,
          offset: lastTextNode.getTextContent().length
        };
      }

      const patternMatchResults = getPatternMatchResultsForCriteria(criteria, scanningContext, elementNode);

      if (patternMatchResults != null) {
        scanningContext.markdownCriteria = criteria;
        scanningContext.patternMatchResults = patternMatchResults; // Perform text transformation here.

        transformTextNodeForMarkdownCriteria(scanningContext, elementNode, createHorizontalRuleNode);
        resetScanningContext(scanningContext);
        const currentTextContent = elementNode.getTextContent();

        if (currentTextContent.length === 0) {
          // Nothing left to convert.
          return;
        }

        if (currentTextContent === textContent) {
          // Nothing was changed by this transformation, so move on to the next crieteria.
          continue;
        } // The text was changed. Perhaps there is another hit for the same criteria.


        textContent = currentTextContent;
        startIndex = i;
        done = false;
        break;
      }
    }
  }
}

function getLastTextNodeInElementNode(elementNode) {
  const children = elementNode.getChildren();
  const countOfChildren = children.length;

  for (let i = countOfChildren - 1; i >= 0; i--) {
    if (lexical.$isTextNode(children[i])) {
      return children[i];
    }
  }

  return null;
}

/**
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 */
function registerMarkdownShortcuts(editor, createHorizontalRuleNode) {
  // The priorTriggerState is compared against the currentTriggerState to determine
  // if the user has performed some typing event that warrants an auto format.
  // For example, typing "#" and then " ", shoud trigger an format.
  // However, given "#A B", where the user delets "A" should not.
  let priorTriggerState = null;
  return editor.registerUpdateListener(({
    tags
  }) => {
    // Examine historic so that we are not running autoformatting within markdown.
    if (tags.has('historic') === false) {
      const currentTriggerState = getTriggerState(editor.getEditorState());
      const scanningContext = currentTriggerState == null ? null : findScanningContext(editor, currentTriggerState, priorTriggerState);

      if (scanningContext != null) {
        updateAutoFormatting(editor, scanningContext, createHorizontalRuleNode);
      }

      priorTriggerState = currentTriggerState;
    } else {
      priorTriggerState = null;
    }
  });
}
function $convertFromMarkdownString(markdownString, editor, createHorizontalRuleNode) {
  if (convertStringToLexical(markdownString) != null) {
    convertMarkdownForElementNodes(editor, createHorizontalRuleNode);
  }
}

exports.$convertFromMarkdownString = $convertFromMarkdownString;
exports.registerMarkdownShortcuts = registerMarkdownShortcuts;
